
#include "inc.h"

#include "minix.h"
#include "irq_arm.h"
#include "bios_arm.h"
#include "div_arm.h"
#include "gbfs.h"
#include "lib.h"

#include "rand.h"
#include "screen.h"
#include "main.h"
#include "lisp.h"

//TODO G[̓VXeG[ɂĂ܂B܂t[YiOOGB

//---------------------------------------------------------------------------
ST_SYMBOL Symbol[SYMBOL_MAX_CNT];
ST_CONS   Cons[CONS_MAX_CNT] EWRAM_BSS;
ST_READ   Read;
ST_LISP   Lisp;

u8  printBuf[PRINT_MAX_BUF];
u16 SymbolCnt;
u16 ConsCnt;


// initialize
//---------------------------------------------------------------------------
EWRAM_CODE void LispInit()
{
	_Memset((u8*)&Symbol,   0x00, sizeof(ST_SYMBOL) * SYMBOL_MAX_CNT);
	_Memset((u8*)&Cons,     0x00, sizeof(ST_CONS)   * CONS_MAX_CNT);
	_Memset((u8*)&Read,     0x00, sizeof(ST_READ));
	_Memset((u8*)&Lisp,     0x00, sizeof(ST_LISP));
	_Memset((u8*)&printBuf, 0x00, PRINT_MAX_BUF);
	SymbolCnt = 0;
	ConsCnt   = 0;

	ST_LISP* p = &Lisp;

	// nil
	cell n    = LispMakeSymbol((u8*)"nil");
	p->topEnv = LispFuncAcons(n, NIL, NIL);

	// symbols
	LispMakeSymbolVar((u8*)"t"      , &p->symT);
	LispMakeSymbolVar((u8*)"*func*" , &p->symFunc);
	LispMakeSymbolVar((u8*)"*sform*", &p->symSform);
	LispMakeSymbolVar((u8*)"macro"  , &p->symMacro);

	// primitive fuctions
	LispMakeFunc((u8*)"car"  , &LispFuncCar  ,   1);
	LispMakeFunc((u8*)"cdr"  , &LispFuncCdr  ,   1);
	LispMakeFunc((u8*)"cons" , &LispFuncCons ,   2);
	LispMakeFunc((u8*)"eq"   , &LispFuncEq   ,   2);
	LispMakeFunc((u8*)"+"    , &LispFuncAdd  ,   2);
	LispMakeFunc((u8*)"-"    , &LispFuncSub  ,   2);
	LispMakeFunc((u8*)"*"    , &LispFuncMul  ,   2);
	LispMakeFunc((u8*)"/"    , &LispFuncDiv  ,   2);
	LispMakeFunc((u8*)"="    , &LispFuncEq   ,   2);
	LispMakeFunc((u8*)">"    , &LispFuncGt   ,   2);
	LispMakeFunc((u8*)"assoc", &LispFuncAssoc,   2);
	LispMakeFunc((u8*)"acons", &LispFuncAcons,   3);
	LispMakeFunc((u8*)"list" , &LispFuncListF,  -1);
	LispMakeFunc((u8*)"eval" , &LispFuncEval , 101);

	// special-forms
	LispMakeSform((u8*)"setq", &LispFuncSetq);
	LispMakeSform((u8*)"cond", &LispFuncCond);

	LispMakeSformVar((u8*)"quote" , &LispFuncQuote , &p->symQuote);
	LispMakeSformVar((u8*)"lambda", &LispFuncLambda, &p->symLambda);
	LispMakeSformVar((u8*)"macro" , &LispFuncMacro , &p->symMacro);
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispMakeSymbol(u8* name)
{
	cell c;

	//o^ς݂̏ꍇÃV{Ԃ܂
	c = LispFindSymbol(name);
	if(c != NULL)
	{
		return  c;
	}

	//V{̐VKo^܂
	if(SymbolCnt >= SYMBOL_MAX_CNT)
	{
		SystemError("Symbol cnt overflow", NULL);
	}

	if( _Strlen(name) >= SYMBOL_MAX_BUF )
	{
		SystemError("Symbol buf overflow", NULL);
	}

	ST_SYMBOL* s = &Symbol[SymbolCnt];

	_Strcpy(s->name, name);
	SymbolCnt++;

	return (cell)s;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFindSymbol(u8* name)
{
	u16 i;

	for(i=0; i<SymbolCnt; i++)
	{
		if( _Strcmp(Symbol[i].name, name) == 0 )
		{
			return (cell)&Symbol[i];
		}
	}

	return NULL;
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispMakeSymbolVar(u8* name, cell* var)
{
	*var = LispMakeSymbol(name);

	Lisp.topEnv = LispFuncAcons(*var, *var, Lisp.topEnv);
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispMakeFunc(u8* name, void* func, s32 args)
{
	cell a = LispFuncCons(INT2CELL(func), INT2CELL(args));
	cell b = LispFuncCons(Lisp.symFunc, a);

	cell s = LispMakeSymbol(name);
	Lisp.topEnv = LispFuncAcons(s, b, Lisp.topEnv);
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispMakeSform(u8* name, void* func)
{
	cell s = LispMakeSymbol(name);
	cell a = LispFuncCons(Lisp.symSform, INT2CELL(func));

	Lisp.topEnv = LispFuncAcons(s, a, Lisp.topEnv);
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispMakeSformVar(u8* name, void* func, cell* var)
{
	*var   = LispMakeSymbol(name);
	cell a = LispFuncCons(Lisp.symSform, INT2CELL(func));

	Lisp.topEnv = LispFuncAcons(*var, a, Lisp.topEnv);
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispMakeListCell(char* s)
{
	LispReadSetBuf((u8*)s);
	return LispReadNext(LispReadGetToken());
}


//primitive functions
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncCar(cell e)
{
	if(e == NIL)
	{
		return  NIL;
	}

	return CAR(e);
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncCdr(cell e)
{
	if(e == NIL)
	{
		return  NIL;
	}

	return CDR(e);
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncCons(cell car, cell cdr)
{
	if(ConsCnt >= CONS_MAX_CNT)
	{
		SystemError("Cons cnt overflow", NULL);
	}

	ST_CONS* c = &Cons[ConsCnt];

	c->car = car;
	c->cdr = cdr;
	ConsCnt++;

	return (cell)c;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncConCat2(cell a, cell b)
{
	if(a == NIL)
	{
		return b;
	}
	if(b == NIL)
	{
		return a;
	}

	cell p = a;
	while( CDR(p) != NIL )
	{
		p = CDR(p);
	}
	CDR(p) = b;

	return a;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncAppend2(cell a, cell b)
{
	if (a == NIL)
	{
		return b;
	}

	return LispFuncCons( CAR(a), LispFuncAppend2(CDR(a), b) );
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncAdd(cell a, cell b)
{
	return INT2CELL( (CELL2INT(a) + CELL2INT(b)) );
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncSub(cell a, cell b)
{
	return INT2CELL( (CELL2INT(a) - CELL2INT(b)) );
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncMul(cell a, cell b)
{
	return INT2CELL( (CELL2INT(a) * CELL2INT(b)) );
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncDiv(cell a, cell b)
{
	return INT2CELL( (s32)_Div(CELL2INT(a), CELL2INT(b)) );
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncEq(cell a, cell b)
{
	return (a == b) ? Lisp.symT : NIL;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncGt(cell a, cell b)
{
	return (CELL2INT(a) > CELL2INT(b)) ? Lisp.symT : NIL;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncAssoc(cell key, cell list)
{
	while(list != NIL)
	{
		if(key == CAAR(list))
		{
			return CAR(list);
		}

		list = CDR(list);
	}

	return NIL;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncAcons(cell key, cell value, cell alist)
{
	return LispFuncCons( LispFuncCons(key, value), alist );
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncPairlis(cell keys, cell values)
{
	cell alist = NIL;

	while(keys != NIL)
	{
		alist = LispFuncAcons(CAR(keys), (values == NIL) ? NIL : CAR(values), alist);

		keys   = CDR(keys);
		values = CDR(values);
	}

	return alist;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncListF(cell list)
{
	return list;
}


// special-form
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncQuote(cell e, cell env)
{
	return CAR(e);
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncSetq(cell e, cell env)
{
	cell v = CAR(e);
	cell s = CADR(e);

	if( LispIsSymbol(v) == FALSE )
	{
		//throw error
		SystemError("setq require symbol", "for 1st argument");
	}

	cell v_env = LispFuncAssoc(v, env);
	if(v_env == NIL)
	{
		v_env = LispFuncAssoc(v, Lisp.topEnv);
	}

	if(v_env != NIL)
	{
		CDR(v_env) = LispFuncEval(s, env); 
		return v;
	}

	if(env == NIL)
	{
		Lisp.topEnv = LispFuncAcons(v, LispFuncEval(s, env), Lisp.topEnv);
		return v;
	}

	SystemError("unbound variable", NULL);
//	throw_error("unbound variable: %s", SYMBOL_NAME(v));
	return NULL;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncCond(cell e, cell env)
{
	while(e != NIL && LispFuncEval(CAAR(e), env) == NIL)
	{
		e = CDR(e);
	}

	return (e != NIL) ? LispFuncEval(CAR(CDAR(e)), env) : NIL;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncLambda(cell e, cell env)
{
	return LispFuncCons(Lisp.symLambda, LispFuncConCat2(e, LispFuncCons(env, NIL)));
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncMacro(cell e, cell env)
{
	return LispFuncCons(Lisp.symMacro, LispFuncConCat2(e, LispFuncCons(env, NIL)));
}


//read
//---------------------------------------------------------------------------
EWRAM_CODE ST_TOKEN* LispReadGetToken()
{
	static ST_TOKEN token;
	static u8 sym[SYMBOL_MAX_BUF];

	u8  c;
	s16 n;
	s16 len;

	do {
		c = LispReadGetChr();

	} while(_IsSpace(c) == TRUE || c == '\t');

	//ʂȕ`FbN܂
	if( _Strchr(TOKEN_SPECIAL_STR, c) != NULL )
	{
		token.kind = c;
		return &token;
	}

	//`FbN܂
	if( _IsDigit(c) == TRUE )
	{
		n = (c - '0');
		c = LispReadGetChr();

		while ( _IsDigit(c) == TRUE )
		{
			n = n * 10 + (c - '0');
			c = LispReadGetChr();
		}
		LispReadUnGetChr();

		token.kind = 'N';
		token.num  = n;
		return &token;
	}

	//}CiXij`FbN܂
	if(c == '-')
	{
		c = LispReadGetChr();

		if( _IsDigit(c) == TRUE )
		{
			LispReadUnGetChr();

			ST_TOKEN* t  = LispReadGetToken();
			t->num      *= -1;
			return t;
		}

		LispReadUnGetChr();
		c = '-';
	}

	//܂ŗꍇAV{ɂȂ܂
	len = 0;
	do {
		sym[len++] = c;
		c = LispReadGetChr();

	} while(! (_Strchr(TOKEN_SPECIAL_STR, c) != NULL || _IsSpace(c) == TRUE || c == '\t') );

	sym[len] = '\0';
	LispReadUnGetChr();

	token.kind = 'S';
	token.name = sym;
	return &token;
}
//---------------------------------------------------------------------------
EWRAM_CODE u8 LispReadGetChr()
{
	ST_READ* p = &Read;
	bool isComment = FALSE;
	u8   c;

	p->pCurPrev = p->pCur;

	for(;;)
	{
		if((u32)p->pCur - (u32)p->pFile > p->size)
		{
			return '\0';
		}

		c = *p->pCur++;

		if(c == CR)
		{
			continue;
		}

		if(c == ';')
		{
			isComment = TRUE;
		}
		else if(c == LF)
		{
			isComment = FALSE;
		}
		else if(isComment == FALSE)
		{
			return c;
		}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispReadUnGetChr()
{
	Read.pCur = Read.pCurPrev;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispReadNext(ST_TOKEN* t)
{
	cell car_e, cdr_e, e;
	cell list = NIL;

	if(t->kind != '(')
	{
		if(t->kind == 'N')
		{
			return INT2CELL(t->num);
		}

		if(t->kind == '\'')
		{
			e = LispReadNext(LispReadGetToken());
			return LispFuncCons(Lisp.symQuote, LispFuncCons(e, NIL));
		}

		return LispMakeSymbol(t->name);
	}

	//(
	for(;;)
	{
		t = LispReadGetToken();

		if(t->kind == ')')
		{
			return list;
		}

		car_e = LispReadNext(t);

		if(t->kind == '.')
		{
			// (a.b)
			t     = LispReadGetToken();
			cdr_e = LispReadNext(t);
			t     = LispReadGetToken();

			if (t->kind != ')')
			{
				SystemError(") required after", NULL);
				//throw_error(") required after '%s'", sprint_e(cdr_e));
			}

			return LispFuncCons(car_e, cdr_e);
		}

		// (a b)
		list = LispFuncConCat2(list, LispFuncCons(car_e, NIL));
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispReadCell()
{
	return LispReadNext(LispReadGetToken());
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispReadFile(u8* pFile, u32 size)
{
	Read.pFile = pFile;
	Read.pCur  = pFile;
	Read.size  = size;
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispReadSetBuf(u8* pBuf)
{
	LispReadFile(pBuf, _Strlen(pBuf));
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispExec()
{
	ST_TOKEN* t = LispReadGetToken();

	while(t->kind != '\0')
	{
		LispPrintBuf(LispFuncEval(LispReadNext(t), NIL));
		t = LispReadGetToken();
	}
}


// print
//---------------------------------------------------------------------------
EWRAM_CODE void LispSprintNext(u8* buf, cell e)
{
	u8 w[32];

	if( _Strlen(buf) >= PRINT_MAX_BUF - 32)
	{
		return;
	}

	if(e == NIL)
	{
		_Strcat(buf, (u8*)"nil");
		return;
	}

	if( LispIsInt(e) == TRUE )
	{
		_Sprintf((char*)w, "%d", CELL2INT(e));
		_Strcat(buf, w);
		return;
	}

	if( LispIsSymbol(e) == TRUE )
	{
		_Strcat(buf, SYMBOL_NAME(e));
		return;
	}

	if( LispIsCons(e) == TRUE )
	{
		_Strcat(buf, (u8*)"(");

		for(;;)
		{
			LispSprintNext(buf, CAR(e));
			e = CDR(e);

			if(e == NIL)
			{
				break;
			}
			else if( (LispIsInt(e) == TRUE) || (LispIsSymbol(e) == TRUE) )
			{
				_Strcat(buf, (u8*)".");
				LispSprintNext(buf, e);
				break;
			}

			_Strcat(buf, (u8*)" ");
		}

		_Strcat(buf, (u8*)")");
		return;
	}

	_Sprintf((char*)w, "*%x*", e);
	_Strcat((u8*)buf, (u8*)w);
}
//---------------------------------------------------------------------------
EWRAM_CODE u8* LispSprintBuf(cell e)
{
	printBuf[0] = '\0';
	LispSprintNext(printBuf, e);

	return printBuf;
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispPrintBuf(cell e)
{
	ScreenImgDrawStr((u8*)LispSprintBuf(e));
	ScreenImgLineFeed();

	ScreenImgUpdate();
}





// eval
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncEval(cell e, cell env)
{
	cell v;

	if( e == NIL || LispIsInt(e) == TRUE )
	{
		return e;
	}

	if( LispIsSymbol(e) == TRUE )
	{
		v = LispFuncAssoc(e, env);

		if(v == NIL)
		{
			v = LispFuncAssoc(e, Lisp.topEnv);
		}

		if(v != NIL)
		{
			return CDR(v);
		}

		SystemError("Undefined symbol", NULL);
		//throw_error("Undefined symbol '%s'", SYMBOL_NAME(e));
	}

	return LispFuncApply(CAR(e), CDR(e), env);
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncEvalList(cell list, cell env)
{
	cell values = NIL;

	while(list != NIL)
	{
		values = LispFuncConCat2(values, LispFuncCons(LispFuncEval(CAR(list), env), NIL));
		list   = CDR(list);
	}

	return values;
}
//---------------------------------------------------------------------------
EWRAM_CODE cell LispFuncApply(cell func, cell args, cell env)
{
	cell fbody = LispFuncEval(func, env);

	if( LispIsCons(fbody) == FALSE )
	{
		SystemError("Not function", NULL);
		//throw_error("Not function %s", sprint_e(func));
	}

	cell ftype = CAR(fbody);

	if(ftype == Lisp.symFunc)
	{
		cell (*func_pt)() = (void*)CELL2INT(CADR(fbody));
		s32  n_func_args  = CELL2INT(CDDR(fbody));

		if(n_func_args > 0)
		{
			LispChkArgCnt(func, _UMod(n_func_args, 100), LispGetlistLen(args));
		}
		switch (n_func_args)
		{
			case -1:
				return (*func_pt)(LispFuncEvalList(args, env));
			case  0:
				return (*func_pt)();
			case  1:
				return (*func_pt)(LispFuncEval(CAR(args), env));
			case  2:
				return (*func_pt)(LispFuncEval(CAR(args), env), LispFuncEval(CADR(args), env));
			case  3:
				return (*func_pt)(LispFuncEval(CAR(args), env), LispFuncEval(CADR(args), env), LispFuncEval(CADDR(args), env));
			case 100:
				return (*func_pt)(env);
			case 101:
				return (*func_pt)(LispFuncEval(CAR(args), env), env);
			case 102:
				return (*func_pt)(LispFuncEval(CAR(args), env), LispFuncEval(CADR(args), env), env);
			case 103:
				return (*func_pt)(LispFuncEval(CAR(args), env), LispFuncEval(CADR(args), env), LispFuncEval(CADDR(args), env), env);
		}
	}

	if(ftype == Lisp.symSform)
	{
		cell (*sform_pt)() =  (void *)CELL2INT(CDR(fbody));
		return (*sform_pt)(args, env);
	}

	if(ftype == Lisp.symLambda)
	{
		cell params      = CADR(fbody);
		cell e           = CADDR(fbody);
		cell closure_env = CADDDR(fbody);
		LispChkArgCnt(func, LispGetlistLen(params), LispGetlistLen(args));

		return LispFuncEval(e, LispFuncAppend2(LispFuncConCat2(LispFuncPairlis(params, LispFuncEvalList(args, env)), closure_env), env));
	}

	if(ftype == Lisp.symMacro)
	{
		cell params      = CADR(fbody);
		cell e           = CADDR(fbody);
		cell closure_env = CADDDR(fbody);
		LispChkArgCnt(func, LispGetlistLen(params), LispGetlistLen(args));

		cell r = LispFuncEval(e, LispFuncAppend2(LispFuncConCat2(LispFuncPairlis(params, args), closure_env), env));
		return LispFuncEval(r, env);
	}

	SystemError("Not function", NULL);
	//throw_error("Not function %s", sprint_e(func));*/
	return NULL;
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispChkArgCnt(cell f, s32 requires, s32 arg_count)
{
	if(arg_count != requires)
	{
		SystemError("Wrong number of arguments\n", NULL);
		//throw_error("Wrong number of arguments '%s' requires %d, but %d", sprint_e(f), requires, arg_count);
	}
}




// macro
//---------------------------------------------------------------------------
EWRAM_CODE bool LispIsInt(cell e)
{
	return (((s32)e & 0x00000001) != 0) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool LispIsSymbol(cell e)
{
	return ((ST_SYMBOL*)e >= &Symbol[0] && (ST_SYMBOL*)e < &Symbol[SymbolCnt]) ? TRUE : FALSE;
}
//---------------------------------------------------------------------------
EWRAM_CODE bool LispIsCons(cell e)
{
	return ((ST_CONS*)e >= &Cons[0] && (ST_CONS*)e < &Cons[ConsCnt]) ? TRUE : FALSE;
}


// utility
//---------------------------------------------------------------------------
EWRAM_CODE u16 LispGetlistLen(cell list)
{
	u16 ret = 0;

	while(list != NIL)
	{
		list = CDR(list);
		ret++;
	}

	return ret;
}


// test tool
//---------------------------------------------------------------------------
EWRAM_CODE void LispTest()
{
	LispTestMakeSymbol();
	LispTestPrint();
	LispTestToken();
	LispTestRead();
	LispTestFunction();
	LispTestSpecialForm();
	LispTestApplyEval();
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispAssertStr(char* expected, char* value, char* msg)
{
	TRACEOUT(" %s -- ", msg);

	if( _Strcmp((u8*)expected, (u8*)value) == 0 )
	{
		TRACEOUT(" OK\n");
	}
	else
	{
		TRACEOUT(" NG value = %s, expected = %s\n", value, expected);
		for(;;){}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispAssertInt(s16 expected, s16 value, char* msg)
{
	TRACEOUT(" %s -- ", msg);

	if(expected == value)
	{
		TRACEOUT(" OK\n");
	}
	else
	{
		TRACEOUT(" NG value = %d, expected = %d\n", value, expected);
		for(;;){}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispAssertLisp(char* expected, cell value, char* msg)
{
	LispAssertStr(expected, (char*)LispSprintBuf((u8*)value), msg);
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispAssertChr(char* expected, u8 value, char* msg)
{
	char s[2];
	s[0] = value;
	s[1] = '\0';

	LispAssertStr(expected, s, msg);
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispAssertBool(bool is, char* msg)
{
	TRACEOUT(" %s -- ", msg);

	if(is == TRUE)
	{
		TRACEOUT(" OK\n");
	}
	else
	{
		TRACEOUT(" NG\n");
		for(;;){}
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispTestMakeSymbol()
{
	cell a = LispMakeSymbol((u8*)"a");
	cell b = LispMakeSymbol((u8*)"b");

	LispAssertStr("a", (char*)SYMBOL_NAME(a), "make symbol - name");
	LispAssertStr("b", (char*)SYMBOL_NAME(b), "make symbol - name");

	LispAssertBool(a == LispMakeSymbol((u8*)"a"), "make symbol - ident");
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispTestPrint()
{
	cell a = LispMakeSymbol((u8*)"a");
	cell b = LispMakeSymbol((u8*)"b");

	LispAssertStr("123", (char*)LispSprintBuf(INT2CELL(123)), "print(number)");
	LispAssertStr("abc", (char*)LispSprintBuf(LispMakeSymbol((u8*)"abc")), "print(symbol)");
	LispAssertStr("(a)", (char*)LispSprintBuf(LispFuncCons(a, NIL)), "print((a))");
	LispAssertStr("(a.b)", (char*)LispSprintBuf(LispFuncCons(a,b)), "print((a.b))");
	LispAssertStr("(a b)", (char*)LispSprintBuf(LispFuncCons(a,LispFuncCons(b, NIL))), "print((a b))");
	LispAssertStr("(a 123 b)", (char*)LispSprintBuf(LispFuncCons(a,LispFuncCons(INT2CELL(123), LispFuncCons(b, NIL)))), "print((a 123 b))");
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispTestToken()
{
	ST_TOKEN* tk;
	LispReadSetBuf((u8*)"(123 abcd\t-56 .-)\n");

	tk = LispReadGetToken();
	LispAssertChr("(", tk->kind, "get_token (");

	tk = LispReadGetToken();
	LispAssertChr("N", tk->kind, "get_token 123 (tk)");
	LispAssertInt(123, tk->num, "get_token 123 (number)");

	tk = LispReadGetToken();
	LispAssertChr("S", tk->kind, "get_token abcd (tk)");
	LispAssertStr("abcd", (char*)tk->name, "get_token abcd (name)");

	tk = LispReadGetToken();
	LispAssertChr("N", tk->kind, "get_token -56 (tk)");
	LispAssertInt(-56, tk->num, "get_token -56 number)");

	tk = LispReadGetToken();
	LispAssertChr(".", tk->kind, "get_token .");

	tk = LispReadGetToken();
	LispAssertChr("S", tk->kind, "get_token - (tk)");
	LispAssertStr("-", (char*)tk->name, "get_token - (name)");

	tk = LispReadGetToken();
	LispAssertChr(")", tk->kind, "get_token )");
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispTestRead()
{
	LispAssertLisp("12345", LispMakeListCell("12345\n"), "read 12345");
	LispAssertLisp("ab+cd", LispMakeListCell("ab+cd\n"), "read ab+cd");
	LispAssertLisp("nil", LispMakeListCell("( )\n"), "read ()");
	LispAssertLisp("(a)", LispMakeListCell("( a )\n"), "read (a)");
	LispAssertLisp("(a.b)", LispMakeListCell("( a . b )\n"), "read (a.b)");
	LispAssertLisp("(a b)", LispMakeListCell("( a  b )\n"), "read (a b)");
	LispAssertLisp("(a b)", LispMakeListCell("( a  b )\n"), "read (a b)");
	LispAssertLisp("(a b c)", LispMakeListCell("( a  b c)\n"), "read (a b c)");
	LispAssertLisp("((a) (b c))", LispMakeListCell("((a)  (b c))\n"), "read ((a) (b c))");
	LispAssertLisp("((a (b (c d))) e)", LispMakeListCell("(( a ( b (c  d) ) ) e)\n"), "read ((a (b (c d))) e)");
	LispAssertLisp("(quote a)", LispMakeListCell("'a\n"), "read 'a");
	LispAssertLisp("(quote (a b c))", LispMakeListCell("'(a b c)\n"), "read '(a b c)");
	LispAssertLisp("((a.1) (b.1))", LispMakeListCell("((a.1)(b.1))"), "read ((a.1) (b.1))");
	LispAssertLisp("(c = n 0)", LispMakeListCell("(c.(= n 0))"), "read (c.(= n 0))");
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispTestFunction()
{
	cell a     = LispMakeSymbol((u8*)"a");
	cell b     = LispMakeSymbol((u8*)"b");
	cell c     = LispMakeSymbol((u8*)"c");
	cell d     = LispMakeSymbol((u8*)"d");
	cell l     = LispFuncCons(a, LispFuncCons(b, NIL));
	cell alist = NIL;

	LispAssertLisp("(a b)", l, "cons(a (cons b nil))");
	LispAssertLisp("(a)", LispFuncConCat2(NIL, LispFuncCons(a, NIL)), "concat2(nil, (a))");
	LispAssertLisp("(a b c)", LispFuncConCat2(l, LispFuncCons(c, NIL)), "concat2((a b), (c))");

	LispAssertLisp("a", LispFuncCar(l), "car((abc))");
	LispAssertLisp("(b c)", LispFuncCdr(l), "cdr((abc))");

	LispAssertInt(3, LispGetlistLen(l), "list_len((a b))");
	LispAssertLisp("(x y a b c)", LispFuncAppend2(LispMakeListCell("(x y)"), l), "append2((x y), (a b c))");

	alist = LispFuncAcons(a, INT2CELL(1), alist);
	alist = LispFuncAcons(b, INT2CELL(2), alist);
	alist = LispFuncAcons(c, INT2CELL(3), alist);

	LispAssertLisp("((c.3) (b.2) (a.1))", alist, "acons()");
	LispAssertLisp("(b.2)", LispFuncAssoc(b, alist), "asooc(ok)");
	LispAssertLisp("nil", LispFuncAssoc(d, alist), "asooc(nil)");
	LispAssertLisp("((b.2) (a.1))", LispFuncPairlis(LispMakeListCell("(a b)"), LispMakeListCell("(1 2)")), "pairlis");
	LispAssertLisp("(11 22 33)", LispFuncListF(LispMakeListCell("(11 22 33)")), "list");

	LispAssertLisp("5", LispFuncAdd(LispMakeListCell("2"), LispMakeListCell("3")), "add()");
	LispAssertLisp("1", LispFuncSub(LispMakeListCell("3"), LispMakeListCell("2")), "subtract()");
	LispAssertLisp("6", LispFuncMul(LispMakeListCell("2"), LispMakeListCell("3")), "multiply()");
	LispAssertLisp("2", LispFuncDiv(LispMakeListCell("10"), LispMakeListCell("4")), "divide()");
	LispAssertLisp("t", LispFuncGt(LispMakeListCell("3"), LispMakeListCell("2")), "gt():t");
	LispAssertLisp("nil", LispFuncGt(LispMakeListCell("2"), LispMakeListCell("3")), "gt():nil");
	LispAssertLisp("t", LispFuncEq(a, a), "eq():t");
	LispAssertLisp("nil", LispFuncEq(a, b), "eq():nil");
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispTestSpecialForm()
{
	cell a   = LispMakeSymbol((u8*)"a");
	cell b   = LispMakeSymbol((u8*)"b");
	cell c   = LispMakeSymbol((u8*)"c");
	cell env = Lisp.topEnv;

	env = LispFuncAcons(a, LispMakeListCell("11"), env);
	env = LispFuncAcons(b, NIL, env);

	LispFuncSetq(LispFuncCons(b, LispFuncCons(a, NIL)), env);
	LispAssertLisp("11", CDR(LispFuncAssoc(b, env)), "setq(b a)");

	LispFuncSetq(LispFuncCons(c, LispFuncCons(INT2CELL(22), NIL)), NIL);
	LispAssertLisp("22", CDR(LispFuncAssoc(c, Lisp.topEnv)), "setq(c 22) top level");

	LispAssertLisp("1", LispFuncCond(LispMakeListCell("(((> 4 3) 1) ((= 4 3) 0) (t -1))"), env), "cond 1");
	LispAssertLisp("0", LispFuncCond(LispMakeListCell("(((> 3 3) 1) ((= 3 3) 0) (t -1))"), env), "cond 2");
	LispAssertLisp("-1", LispFuncCond(LispMakeListCell("(((> 1 3) 1) ((= 1 3) 0) (t -1))"), env), "cond 3");

	LispAssertLisp("(lambda (x y) (+ x y) ((x.1) (y.2)))", LispFuncLambda(LispMakeListCell("((x y) (+ x y))"), LispMakeListCell("((x.1) (y.2))")), "lambda");
	LispAssertLisp(
		"(macro (c th el) (list (quote cond) (list c th) (list t el)) ((c = n 0) (th.1) (el.5)))",
		LispFuncMacro(LispMakeListCell("((c th el) (list 'cond (list c th) (list t el)))"),
		LispMakeListCell("((c.(= n 0)) (th.1) (el.5))")), "macro");
}
//---------------------------------------------------------------------------
EWRAM_CODE void LispTestApplyEval()
{
	cell a, b;
	LispMakeSymbolVar((u8*)"a", &a);
	LispMakeSymbolVar((u8*)"b", &b);
	cell l = LispFuncCons(a, LispFuncCons(b, NIL));
	cell env = Lisp.topEnv; 

	LispAssertLisp("(a.b)", LispFuncApply(LispMakeListCell("cons"), l, env), "apply(func)");
	LispAssertLisp("(a b)", LispFuncApply(LispMakeListCell("quote"), LispFuncCons(l, NIL), env), "apply(sform)");
	LispAssertLisp("3",
		LispFuncApply(LispFuncLambda(LispMakeListCell("((x y) (+ x y))"), env), 
		LispMakeListCell("(1 2)"), env), "apply(lambda)");

  LispAssertLisp("8",
		LispFuncApply(LispFuncLambda(LispMakeListCell("((x) (+ x n))"),
		LispFuncCons(LispMakeListCell("(n.5)"), env)),
		LispMakeListCell("(3)"), env), "apply(lambda+closure)");


	char *mbody = "((c th el) (list 'cond (list c th) (list t el)))";
	LispAssertLisp("5", LispFuncApply(LispFuncMacro(LispMakeListCell(mbody), env),
	LispMakeListCell("((= n 0) 1 5)"), LispFuncCons(LispMakeListCell("(n.2)"), env)), "macro 1");
	LispAssertLisp("1", LispFuncApply(LispFuncMacro(LispMakeListCell(mbody), env),
		LispMakeListCell("((= n 0) 1 5)"), 
		LispFuncCons(LispMakeListCell("(n.0)"), env)), "macro 2");

	LispAssertLisp("nil", LispFuncEval(NIL, env), "eval(nil)");
	LispAssertLisp("11", LispFuncEval(INT2CELL(11), env), "eval(number)");
	LispAssertLisp("a", LispFuncEval(a, env), "eval(symbol value)");
	LispAssertLisp("2", LispFuncEval(a, LispMakeListCell("((b.1)(a.2))")), "eval(bind value)");
}
