
#include "inc.h"

#include "minix.h"
#include "irq_arm.h"
#include "bios_arm.h"
#include "div_arm.h"
#include "mml_arm.h"
#include "ad_arm.h"
#include "gbfs.h"
#include "lib.h"
#include "rand.h"

#include "screen.h"
#include "haribote.h"
#include "lisp.h"

#include "main.h"


//#include "header\font\BGFont.h"


//Mode3
//---------------------------------------------------------------------------
u8 State;
u8 selectMax;
u8 selectCur;

//---------------------------------------------------------------------------
EWRAM_CODE int main()
{
	State = STATE_INIT;

	for(;;)
	{
		switch(State)
		{
			case STATE_INIT:   StateInit();   break;
			case STATE_RESET:  StateReset();  break;
			case STATE_SELECT: StateSelect(); break;
			case STATE_EXEC:   StateExec();   break;

			default:
				SystemError("[Err] main State", NULL);
				break;
		}
	}

	return 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateInit()
{
	InitWS();
	InitLib();
	InitKey();
	InitIRQ();
	InitFont();
	InitSprite();

//	AdInit();
//	MmlInit();
	ScreenImgInit();

	SetMode(MODE_3 | BG2_ENABLE | OBJ_ENABLE | OBJ_MAP_2D);
	InitGBFS();

	State = STATE_RESET;
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitWS()
{
	REG_WSCNT = 0x4317;
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitLib()
{
	LibInit();
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitKey()
{
	KeyInit();
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitIRQ()
{
	REG_INTERUPT = (u32)IRQUserHandler;
	REG_IE       = IRQ_BIT_TIMER2;

	REG_TM2D     = 65536 - 2621;			// (((16*1024*1024) / 64) * 10) / 1000
	REG_TM2CNT   = TM_ENABLE | TM_FREQ_PER_64 | TM_USEIRQ;

	REG_IME      = IRQ_MASTER_ON;
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitFont()
{
	ST_FONT_SRC half;		//p

	half.pDat   = (u8*)&mplus_s10rBitmap;
	half.pSheet = (u16*)&mplus_sfnt_txt;
	half.imgCx  = MPLUS_S10R_IMG_CX;
	half.cnt    = MPLUS_S10R_FONT_CNT;
	half.cx     = MPLUS_S10R_FONT_CX;
	half.cy     = MPLUS_S10R_FONT_CY;

	Mode3SetFont(NULL, &half);
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitSprite()
{
	SpriteInit();

	SpriteSetPalData16((u16*)&sprPal);
	SpriteSetData((u16*)&sprTiles, sprTilesLen);
}
//---------------------------------------------------------------------------
EWRAM_CODE void InitGBFS()
{
	if( GBFSInit() == FALSE )
	{
		SystemError("[Err] GBFS Not Found.", NULL);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateReset()
{
	SelectInit();
	ScreenCurInit();
	LispInit();

//	LispTest();

	State = STATE_SELECT;
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateSelect()
{
/*
	SelectDrawMenu();
	ScreenImgUpdate();

	SelectProcMenu();

	ScreenImgClear();
	ScreenImgUpdate();
*/
	u8* pDat;
	u32 size;


	pDat = GBFSGetFilePointer("base.lisp");
	size = GBFSGetFileSize();
	LispReadFile(pDat, size);
	LispExec();

	pDat = GBFSGetFilePointer("fizzbuzz.lisp");
	size = GBFSGetFileSize();
	LispReadFile(pDat, size);
	LispExec();

	State = STATE_EXEC;
}
//---------------------------------------------------------------------------
EWRAM_CODE void StateExec()
{
	LispReadSetBuf((u8*)"(FizzBuzz)");
	LispExec();

	for(;;)
	{
		WaitForVsync();
	}

	State = STATE_SELECT;
}




//---------------------------------------------------------------------------
EWRAM_CODE void SelectInit()
{
	selectMax = GBFSGetFileCnt();
	selectCur = 0;
}
//---------------------------------------------------------------------------
EWRAM_CODE void SelectDrawMenu()
{
	u16 i;

	for(i=0; i<selectMax; i++)
	{
		GBFSGetFilePointer2(i);
		Mode3DrawFontStr(16, MPLUS_S10R_FONT_CX * i, GBFSGetFileName(), COLOR_WHITE);
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void SelectProcMenu()
{
	u16 key;
	ScreenCurMove(0, selectCur * MPLUS_S10R_FONT_CY);

	for(;;)
	{
		WaitForVsync();
		key = KeyGet();

		if(key == KEY_NONE)
		{
			continue;
		}

		if(key & KEY_A)
		{
			KeyWait3();		//͂ɂȂ܂Ń[v܂
			break;
		}
		else if(key & KEY_UP)
		{
			if(selectCur == 0) continue;
			selectCur--;

			ScreenCurMove(0, selectCur * MPLUS_S10R_FONT_CY);
		}
		else if(key & KEY_DOWN)
		{
			if(selectCur+1 >= selectMax) continue;
			selectCur++;

			ScreenCurMove(0, selectCur * MPLUS_S10R_FONT_CY);
		}
	}

	ScreenCurMove(0, SCREEN_CY);
}




//---------------------------------------------------------------------------
EWRAM_CODE void SystemError(char* msg1, char* msg2)
{
	u16 cy = 11;

	ScreenImgFadeOut(0);
	ScreenImgClear();
	ScreenImgUpdate();

	if(msg1 != NULL) Mode3DrawFontStr(0,  0, (u8*)msg1, COLOR_WHITE);
	if(msg2 != NULL) Mode3DrawFontStr(0, cy, (u8*)msg2, COLOR_WHITE);

	ScreenImgUpdate();
	for(;;){}
}
