
#include "inc.h"

#include "bios_arm.h"
#include "gbfs.h"
#include "lib.h"
#include "main.h"
#include "minix.h"

#include "screen.h"

//---------------------------------------------------------------------------
//lib.c
extern ST_FONT  Font;


ST_SCREEN_IMG  ScreenImg EWRAM_BSS;
ST_CURSOR      Cursor;

u32 dummy ALIGN(4);
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenImgInit()
{
	_Memset((u8*)&ScreenImg, 0x00, sizeof(ST_SCREEN_IMG));
	dummy = 0;

	LibMode3SetScreenBuffer((u16*)&ScreenImg.buf);
	ScreenImgClear();
	ScreenImgUpdate();
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenImgLoad(u16* pData)
{
	Mode3DrawImage(0, 0, SCREEN_CX, SCREEN_CY, pData);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenImgClear()
{
	CpuSet((u32)&dummy, (u32)ScreenImg.buf, ((SCREEN_CX * SCREEN_CY * 2) / 4) | CPUSET_32BIT | CPUSET_SRC_FIX);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenImgUpdate()
{
	WaitForVsync();
	CpuSet((u32)ScreenImg.buf, (u32)VideoBuffer, ((SCREEN_CX * SCREEN_CY * 2) / 4) | CPUSET_32BIT);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenImgFadeIn(u16 mode, u16 min, u16 max, u16 wait)
{
	Mode3DrawBlend(mode, min, max);
	Mode3DrawFadeIn(wait);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenImgFadeOut(u16 wait)
{
	Mode3DrawFadeOut(wait);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenImgDrawStr(u8* str)
{
	ST_FONT* f = &Font;
	ST_SCREEN_IMG* p = &ScreenImg;

	u16 i;
	for(i=0; i<_Strlen(str); i++)
	{
		if(p->x + f->half.cx >= SCREEN_CX)
		{
			ScreenImgLineFeed();
		}


		ScreenImgDrawChr(p->x, p->y, str[i]);
		p->x += f->half.cx;
	}
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenImgDrawChr(u16 x, u16 y, u8 chr)
{
	Mode3DrawFontHalfChr(x, y, chr, COLOR_WHITE);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenImgLineFeed()
{
	ST_FONT* f = &Font;
	ST_SCREEN_IMG* p = &ScreenImg;

	p->x  = 0;
	p->y += f->half.cy;

	if(p->y + f->half.cy >= SCREEN_CY)
	{
		ScreenImgClear();
//		p->y = 0;	//TODO
	}
}



//---------------------------------------------------------------------------
EWRAM_CODE void ScreenSprInit()
{
	CpuSet((u32)&dummy, (u32)OAMmem, ((sizeof(ST_OBJ) * 128) / 4) | CPUSET_32BIT | CPUSET_SRC_FIX);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenSprUpdate()
{

}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenCurInit()
{
	ST_CURSOR* p = &Cursor;

	p->x         = 0;
	p->y         = SCREEN_CY;
	p->sprNo     = 0;
	p->tileNo    = CURSOR_SPRITE_NO;

	SpriteSetSize(p->sprNo, SP_SIZE_16, SP_SQUARE, SP_COLOR_16);
	SpriteSetChr (p->sprNo, p->tileNo);
}
//---------------------------------------------------------------------------
EWRAM_CODE void ScreenCurMove(u16 x, u16 y)
{
	ST_CURSOR* p = &Cursor;

	p->x = x;
	p->y = y;

	SpriteMove(p->sprNo, p->x, p->y);
}
